@@ -105,6 +105,9 @@ static hm_t *reduce_dense_row_by_known_pivots_sparse_ff_8(
         hm_t * const * const pivs,
         const hi_t dpiv,    /* pivot of dense row at the beginning */
         const hm_t tmp_pos, /* position of new coeffs array in tmpcf */
+        const len_t mh,     /* multiplier hash for tracing */
+        const len_t bi,     /* basis index of generating element */
+        const len_t tr,     /* trace data? */
         const uint32_t fc
         )
 {
@@ -117,6 +120,12 @@ static hm_t *reduce_dense_row_by_known_pivots_sparse_ff_8(
     const len_t ncl             = mat->ncl;
     cf8_t * const * const mcf  = mat->cf_8;
 
+    rba_t *rba;
+    if (tr > 0) {
+        rba = mat->rba[tmp_pos];
+    } else {
+        rba = NULL;
+    }
     k = 0;
     for (i = dpiv; i < ncols; ++i) {
         if (dr[i] != 0) {
@@ -137,6 +146,10 @@ static hm_t *reduce_dense_row_by_known_pivots_sparse_ff_8(
         dts   = pivs[i];
         if (i < ncl) {
             cfs   = bs->cf_8[dts[COEFFS]];
+            /* set corresponding bit of reducer in reducer bit array */
+            if (tr > 0) {
+                rba[i/32] |= 1U << (i % 32);
+            }
         } else {
             cfs   = mcf[dts[COEFFS]];
         }
@@ -169,6 +182,8 @@ static hm_t *reduce_dense_row_by_known_pivots_sparse_ff_8(
             j++;
         }
     }
+    row[BINDEX]   = bi;
+    row[MULT]     = mh;
     row[COEFFS]   = tmp_pos;
     row[PRELOOP]  = j % UNROLL;
     row[LENGTH]   = j;
@@ -485,7 +500,7 @@ static cf8_t *reduce_dense_row_by_dense_new_pivots_ff_8(
 static void probabilistic_sparse_reduced_echelon_form_ff_8(
         mat_t *mat,
         const bs_t * const bs,
-        const stat_t * const st
+        md_t *st
         )
 {
     len_t i = 0, j, k, l, m;
@@ -576,7 +591,7 @@ static void probabilistic_sparse_reduced_echelon_form_ff_8(
                     free(npiv);
                     npiv  = NULL;
                     npiv  = reduce_dense_row_by_known_pivots_sparse_ff_8(
-                            drl, mat, bs, pivs, sc, cfp, st->fc);
+                            drl, mat, bs, pivs, sc, cfp, 0, 0, 0, st->fc);
                     if (!npiv) {
                         bctr  = nrbl;
                         break;
@@ -625,6 +640,8 @@ static void probabilistic_sparse_reduced_echelon_form_ff_8(
             memset(dr, 0, (unsigned long)ncols * sizeof(int64_t));
             cfs = mat->cf_8[pivs[k][COEFFS]];
             cfp = pivs[k][COEFFS];
+            const len_t bi  = pivs[k][BINDEX];
+            const len_t mh  = pivs[k][MULT];
             const len_t os  = pivs[k][PRELOOP];
             const len_t len = pivs[k][LENGTH];
             const hm_t * const ds = pivs[k] + OFFSET;
@@ -643,7 +660,7 @@ static void probabilistic_sparse_reduced_echelon_form_ff_8(
             pivs[k] = NULL;
             pivs[k] = mat->tr[npivs++] =
                 reduce_dense_row_by_known_pivots_sparse_ff_8(
-                        dr, mat, bs, pivs, sc, cfp, st->fc);
+                        dr, mat, bs, pivs, sc, cfp, mh, bi, 0, st->fc);
         }
     }
     free(mat->rr);
@@ -656,13 +673,13 @@ static void probabilistic_sparse_reduced_echelon_form_ff_8(
     dr  = NULL;
 
     mat->tr = realloc(mat->tr, (unsigned long)npivs * sizeof(hi_t *));
-    mat->np = mat->nr = mat->sz = npivs;
+    st->np = mat->np = mat->nr = mat->sz = npivs;
 }
 
 static int exact_application_sparse_reduced_echelon_form_ff_8(
         mat_t *mat,
         const bs_t * const bs,
-        const stat_t * const st
+        md_t *st
         )
 {
     len_t i = 0, j, k;
@@ -693,6 +710,8 @@ static int exact_application_sparse_reduced_echelon_form_ff_8(
             int64_t *drl    = dr + (omp_get_thread_num() * ncols);
             hm_t *npiv      = upivs[i];
             cf8_t *cfs      = bs->cf_8[npiv[COEFFS]];
+            const len_t bi  = npiv[BINDEX];
+            const len_t mh  = npiv[MULT];
             const len_t os  = npiv[PRELOOP];
             const len_t len = npiv[LENGTH];
             const hm_t * const ds = npiv + OFFSET;
@@ -713,7 +732,7 @@ static int exact_application_sparse_reduced_echelon_form_ff_8(
                 free(npiv);
                 free(cfs);
                 npiv  = mat->tr[i]  = reduce_dense_row_by_known_pivots_sparse_ff_8(
-                        drl, mat, bs, pivs, sc, i, st->fc);
+                        drl, mat, bs, pivs, sc, i, mh, bi, 0, st->fc);
                 if (!npiv) {
                     fprintf(stderr, "Unlucky prime detected, row reduced to zero.");
                     flag = 0;
@@ -756,6 +775,8 @@ static int exact_application_sparse_reduced_echelon_form_ff_8(
             memset(dr, 0, (unsigned long)ncols * sizeof(int64_t));
             cfs = mat->cf_8[pivs[k][COEFFS]];
             cf_array_pos    = pivs[k][COEFFS];
+            const len_t bi  = pivs[k][BINDEX];
+            const len_t mh  = pivs[k][MULT];
             const len_t os  = pivs[k][PRELOOP];
             const len_t len = pivs[k][LENGTH];
             const hm_t * const ds = pivs[k] + OFFSET;
@@ -774,7 +795,7 @@ static int exact_application_sparse_reduced_echelon_form_ff_8(
             pivs[k] = NULL;
             pivs[k] = mat->tr[npivs++] =
                 reduce_dense_row_by_known_pivots_sparse_ff_8(
-                        dr, mat, bs, pivs, sc, cf_array_pos, st->fc);
+                        dr, mat, bs, pivs, sc, cf_array_pos, mh, bi, 0, st->fc);
         }
     }
     free(pivs);
@@ -783,7 +804,7 @@ static int exact_application_sparse_reduced_echelon_form_ff_8(
     dr  = NULL;
 
     mat->tr = realloc(mat->tr, (unsigned long)npivs * sizeof(hi_t *));
-    mat->np = mat->nr = mat->sz = npivs;
+    st->np = mat->np = mat->nr = mat->sz = npivs;
 
     return 0;
 }
@@ -792,7 +813,7 @@ static void exact_trace_sparse_reduced_echelon_form_ff_8(
         trace_t *trace,
         mat_t *mat,
         const bs_t * const bs,
-        const stat_t * const st
+        md_t *st
         )
 {
     len_t i = 0, j, k;
@@ -885,6 +906,8 @@ static void exact_trace_sparse_reduced_echelon_form_ff_8(
             memset(dr, 0, (unsigned long)ncols * sizeof(int64_t));
             cfs = mat->cf_8[pivs[k][COEFFS]];
             cf_array_pos    = pivs[k][COEFFS];
+            const len_t bi  = pivs[k][BINDEX];
+            const len_t mh  = pivs[k][MULT];
             const len_t os  = pivs[k][PRELOOP];
             const len_t len = pivs[k][LENGTH];
             const hm_t * const ds = pivs[k] + OFFSET;
@@ -903,7 +926,7 @@ static void exact_trace_sparse_reduced_echelon_form_ff_8(
             pivs[k] = NULL;
             pivs[k] = mat->tr[npivs++] =
                 reduce_dense_row_by_known_pivots_sparse_ff_8(
-                        dr, mat, bs, pivs, sc, cf_array_pos, st->fc);
+                        dr, mat, bs, pivs, sc, cf_array_pos, mh, bi, 0, st->fc);
         }
     }
     free(pivs);
@@ -912,13 +935,14 @@ static void exact_trace_sparse_reduced_echelon_form_ff_8(
     dr  = NULL;
 
     mat->tr = realloc(mat->tr, (unsigned long)npivs * sizeof(hi_t *));
-    mat->np = mat->nr = mat->sz = npivs;
+    st->np = mat->np = mat->nr = mat->sz = npivs;
 }
 
 static void exact_sparse_reduced_echelon_form_ff_8(
         mat_t *mat,
+        const bs_t * const tbr,
         const bs_t * const bs,
-        const stat_t * const st
+        md_t *st
         )
 {
     len_t i = 0, j, k;
@@ -929,6 +953,8 @@ static void exact_sparse_reduced_echelon_form_ff_8(
     const len_t ncr   = mat->ncr;
     const len_t ncl   = mat->ncl;
 
+    len_t bad_prime = 0;
+
     /* we fill in all known lead terms in pivs */
     hm_t **pivs   = (hm_t **)calloc((unsigned long)ncols, sizeof(hm_t *));
     memcpy(pivs, mat->rr, (unsigned long)mat->nru * sizeof(hm_t *));
@@ -944,44 +970,78 @@ static void exact_sparse_reduced_echelon_form_ff_8(
     private(i, j, k, sc) \
     schedule(dynamic)
     for (i = 0; i < nrl; ++i) {
-        int64_t *drl    = dr + (omp_get_thread_num() * ncols);
-        hm_t *npiv      = upivs[i];
-        cf8_t *cfs      = bs->cf_8[npiv[COEFFS]];
-        const len_t os  = npiv[PRELOOP];
-        const len_t len = npiv[LENGTH];
-        const hm_t * const ds = npiv + OFFSET;
-        k = 0;
-        memset(drl, 0, (unsigned long)ncols * sizeof(int64_t));
-        for (j = 0; j < os; ++j) {
-            drl[ds[j]]  = (int64_t)cfs[j];
-        }
-        for (; j < len; j += UNROLL) {
-            drl[ds[j]]    = (int64_t)cfs[j];
-            drl[ds[j+1]]  = (int64_t)cfs[j+1];
-            drl[ds[j+2]]  = (int64_t)cfs[j+2];
-            drl[ds[j+3]]  = (int64_t)cfs[j+3];
-        }
-        cfs = NULL;
-        do {
-            sc  = npiv[OFFSET];
-            free(npiv);
-            free(cfs);
-            npiv  = reduce_dense_row_by_known_pivots_sparse_ff_8(
-                    drl, mat, bs, pivs, sc, i, st->fc);
-            if (!npiv) {
-                break;
+        if (bad_prime == 0) {
+            int64_t *drl  = dr + (omp_get_thread_num() * ncols);
+            hm_t *npiv      = upivs[i];
+            cf8_t *cfs      = tbr->cf_8[npiv[COEFFS]];
+            const len_t os  = npiv[PRELOOP];
+            const len_t len = npiv[LENGTH];
+            const len_t bi  = npiv[BINDEX];
+            const len_t mh  = npiv[MULT];
+            const hm_t * const ds = npiv + OFFSET;
+            k = 0;
+            memset(drl, 0, (unsigned long)ncols * sizeof(int64_t));
+            for (j = 0; j < os; ++j) {
+                drl[ds[j]]  = (int64_t)cfs[j];
             }
-            /* normalize coefficient array
-             * NOTE: this has to be done here, otherwise the reduction may
-             * lead to wrong results in a parallel computation since other
-             * threads might directly use the new pivot once it is synced. */
-            if (mat->cf_8[npiv[COEFFS]][0] != 1) {
-                normalize_sparse_matrix_row_ff_8(
-                        mat->cf_8[npiv[COEFFS]], npiv[PRELOOP], npiv[LENGTH], st->fc);
+            for (; j < len; j += UNROLL) {
+                drl[ds[j]]    = (int64_t)cfs[j];
+                drl[ds[j+1]]  = (int64_t)cfs[j+1];
+                drl[ds[j+2]]  = (int64_t)cfs[j+2];
+                drl[ds[j+3]]  = (int64_t)cfs[j+3];
             }
-            k   = __sync_bool_compare_and_swap(&pivs[npiv[OFFSET]], NULL, npiv);
-            cfs = mat->cf_8[npiv[COEFFS]];
-        } while (!k);
+            cfs = NULL;
+            do {
+                sc  = npiv[OFFSET];
+                free(npiv);
+                free(cfs);
+                npiv  = mat->tr[i] = reduce_dense_row_by_known_pivots_sparse_ff_8(
+                        drl, mat, bs, pivs, sc, i, mh, bi, st->trace_level == LEARN_TRACER, st->fc);
+                if (st->nf > 0) {
+                    if (!npiv) {
+                        mat->tr[i]  = NULL;
+                        break;
+                    }
+                    mat->tr[i]  = npiv;
+                    cfs = mat->cf_8[npiv[COEFFS]];
+                    break;
+                } else {
+                    if (!npiv) {
+                        if (st->trace_level == APPLY_TRACER) {
+                            bad_prime = 1;
+                        }
+                        break;
+                    }
+                    /* normalize coefficient array
+                     * NOTE: this has to be done here, otherwise the reduction may
+                     * lead to wrong results in a parallel computation since other
+                     * threads might directly use the new pivot once it is synced. */
+                    if (mat->cf_8[npiv[COEFFS]][0] != 1) {
+                        normalize_sparse_matrix_row_ff_8(
+                                mat->cf_8[npiv[COEFFS]], npiv[PRELOOP], npiv[LENGTH], st->fc);
+                    }
+                    k   = __sync_bool_compare_and_swap(&pivs[npiv[OFFSET]], NULL, npiv);
+                    cfs = mat->cf_8[npiv[COEFFS]];
+                }
+            } while (!k);
+        }
+    }
+
+    if (bad_prime == 1) {
+        for (i = 0; i < ncl+ncr; ++i) {
+            free(pivs[i]);
+            pivs[i] = NULL;
+        }
+        mat->np = 0;
+        if (st->info_level > 0) {
+            fprintf(stderr, "Zero reduction while applying tracer, bad prime.\n");
+        }
+        return;
+    }
+
+    /* construct the trace */
+    if (st->trace_level == LEARN_TRACER) {
+        construct_trace(st->tr, mat);
     }
 
     /* we do not need the old pivots anymore */
@@ -992,52 +1052,57 @@ static void exact_sparse_reduced_echelon_form_ff_8(
 
     len_t npivs = 0; /* number of new pivots */
 
-    dr      = realloc(dr, (unsigned long)ncols * sizeof(uint64_t));
-    mat->tr = realloc(mat->tr, (unsigned long)ncr * sizeof(hm_t *));
+    if (st->nf == 0) {
+        dr      = realloc(dr, (unsigned long)ncols * sizeof(int64_t));
+        mat->tr = realloc(mat->tr, (unsigned long)ncr * sizeof(hm_t *));
 
-    /* interreduce new pivots */
-    cf8_t *cfs;
-    hm_t cf_array_pos;
-    for (i = 0; i < ncr; ++i) {
-        k = ncols-1-i;
-        if (pivs[k]) {
-            memset(dr, 0, (unsigned long)ncols * sizeof(int64_t));
-            cfs = mat->cf_8[pivs[k][COEFFS]];
-            cf_array_pos    = pivs[k][COEFFS];
-            const len_t os  = pivs[k][PRELOOP];
-            const len_t len = pivs[k][LENGTH];
-            const hm_t * const ds = pivs[k] + OFFSET;
-            sc  = ds[0];
-            for (j = 0; j < os; ++j) {
-                dr[ds[j]] = (int64_t)cfs[j];
+        /* interreduce new pivots */
+        cf8_t *cfs;
+        hm_t cf_array_pos;
+        for (i = 0; i < ncr; ++i) {
+            k = ncols-1-i;
+            if (pivs[k]) {
+                memset(dr, 0, (unsigned long)ncols * sizeof(int64_t));
+                cfs = mat->cf_8[pivs[k][COEFFS]];
+                cf_array_pos    = pivs[k][COEFFS];
+                const len_t os  = pivs[k][PRELOOP];
+                const len_t len = pivs[k][LENGTH];
+                const len_t bi  = pivs[k][BINDEX];
+                const len_t mh  = pivs[k][MULT];
+                const hm_t * const ds = pivs[k] + OFFSET;
+                sc  = ds[0];
+                for (j = 0; j < os; ++j) {
+                    dr[ds[j]] = (int64_t)cfs[j];
+                }
+                for (; j < len; j += UNROLL) {
+                    dr[ds[j]]    = (int64_t)cfs[j];
+                    dr[ds[j+1]]  = (int64_t)cfs[j+1];
+                    dr[ds[j+2]]  = (int64_t)cfs[j+2];
+                    dr[ds[j+3]]  = (int64_t)cfs[j+3];
+                }
+                free(pivs[k]);
+                free(cfs);
+                pivs[k] = NULL;
+                pivs[k] = mat->tr[npivs++] =
+                    reduce_dense_row_by_known_pivots_sparse_ff_8(
+                        dr, mat, bs, pivs, sc, cf_array_pos, mh, bi, 0, st->fc);
             }
-            for (; j < len; j += UNROLL) {
-                dr[ds[j]]    = (int64_t)cfs[j];
-                dr[ds[j+1]]  = (int64_t)cfs[j+1];
-                dr[ds[j+2]]  = (int64_t)cfs[j+2];
-                dr[ds[j+3]]  = (int64_t)cfs[j+3];
-            }
-            free(pivs[k]);
-            free(cfs);
-            pivs[k] = NULL;
-            pivs[k] = mat->tr[npivs++] =
-                reduce_dense_row_by_known_pivots_sparse_ff_8(
-                        dr, mat, bs, pivs, sc, cf_array_pos, st->fc);
         }
+        mat->tr = realloc(mat->tr, (unsigned long)npivs * sizeof(hi_t *));
+        st->np = mat->np = mat->nr = mat->sz = npivs;
+    } else {
+        st->np = mat->np = mat->nr = mat->sz = nrl;
     }
     free(pivs);
     pivs  = NULL;
     free(dr);
     dr  = NULL;
-
-    mat->tr = realloc(mat->tr, (unsigned long)npivs * sizeof(hi_t *));
-    mat->np = mat->nr = mat->sz = npivs;
 }
 
 static cf8_t **sparse_AB_CD_linear_algebra_ff_8(
         mat_t *mat,
         const bs_t * bs,
-        const stat_t * const st
+        md_t *st
         )
 {
     len_t i = 0, j;
@@ -1111,7 +1176,7 @@ static cf8_t **sparse_AB_CD_linear_algebra_ff_8(
         free(drs);
         drs = NULL;
     }
-    mat->np = npivs;
+    st->np = mat->np = npivs;
 
     return drs;
 }
@@ -1155,7 +1220,7 @@ static cf8_t **interreduce_dense_matrix_ff_8(
 static cf8_t **exact_dense_linear_algebra_ff_8(
         cf8_t **dm,
         mat_t *mat,
-        const stat_t * const st
+        md_t *st
         )
 {
     len_t i, j, k, l, npivs;
@@ -1256,7 +1321,7 @@ static cf8_t **exact_dense_linear_algebra_ff_8(
             npivs++;
         }
     }
-    mat->np = npivs;
+    st->np = mat->np = npivs;
 
     free(tbr);
     free(dr);
@@ -1267,7 +1332,7 @@ static cf8_t **exact_dense_linear_algebra_ff_8(
 static cf8_t **probabilistic_dense_linear_algebra_ff_8(
         cf8_t **dm,
         mat_t *mat,
-        const stat_t * const st
+        md_t *st
         )
 {
     len_t i, j, k, l, m, npivs;
@@ -1413,7 +1478,7 @@ static cf8_t **probabilistic_dense_linear_algebra_ff_8(
             npivs++;
         }
     }
-    mat->np = npivs;
+    st->np = mat->np = npivs;
 
     free(mul);
     free(tbr);
@@ -1425,7 +1490,7 @@ static cf8_t **probabilistic_dense_linear_algebra_ff_8(
 static cf8_t **probabilistic_sparse_dense_echelon_form_ff_8(
         mat_t *mat,
         const bs_t * const bs,
-        const stat_t * const st
+        md_t *st
         )
 {
     len_t i = 0, j, k, l, m, npivs;
@@ -1550,7 +1615,7 @@ static cf8_t **probabilistic_sparse_dense_echelon_form_ff_8(
             npivs++;
         }
     }
-    mat->np = npivs;
+    st->np = mat->np = npivs;
 
 
     for (i = 0; i < nru; ++i) {
@@ -1647,8 +1712,9 @@ static void convert_to_sparse_matrix_rows_ff_8(
  * linear algebra */
 static void probabilistic_sparse_linear_algebra_ff_8(
         mat_t *mat,
+        const bs_t * const tbr,
         const bs_t * const bs,
-        stat_t *st
+        md_t *st
         )
 {
     /* timings */
@@ -1670,7 +1736,7 @@ static void probabilistic_sparse_linear_algebra_ff_8(
 
     st->num_zerored += (mat->nrl - mat->np);
     if (st->info_level > 1) {
-        printf("%7d new %7d zero", mat->np, mat->nrl - mat->np);
+        printf("%9d new %7d zero", mat->np, mat->nrl - mat->np);
         fflush(stdout);
     }
 }
@@ -1678,7 +1744,7 @@ static void probabilistic_sparse_linear_algebra_ff_8(
 static int exact_application_sparse_linear_algebra_ff_8(
         mat_t *mat,
         const bs_t * const bs,
-        stat_t *st
+        md_t *st
         )
 {
     /* timings */
@@ -1702,7 +1768,7 @@ static int exact_application_sparse_linear_algebra_ff_8(
 
     st->num_zerored += (mat->nrl - mat->np);
     if (st->info_level > 1) {
-        printf("%7d new %7d zero", mat->np, mat->nrl - mat->np);
+        printf("%9d new %7d zero", mat->np, mat->nrl - mat->np);
         fflush(stdout);
     }
 
@@ -1713,7 +1779,7 @@ static void exact_trace_sparse_linear_algebra_ff_8(
         trace_t *trace,
         mat_t *mat,
         const bs_t * const bs,
-        stat_t *st
+        md_t *st
         )
 {
     /* timings */
@@ -1735,15 +1801,16 @@ static void exact_trace_sparse_linear_algebra_ff_8(
 
     st->num_zerored += (mat->nrl - mat->np);
     if (st->info_level > 1) {
-        printf("%7d new %7d zero", mat->np, mat->nrl - mat->np);
+        printf("%9d new %7d zero", mat->np, mat->nrl - mat->np);
         fflush(stdout);
     }
 }
 
 static void exact_sparse_linear_algebra_ff_8(
         mat_t *mat,
+        const bs_t * const tbr,
         const bs_t * const bs,
-        stat_t *st
+        md_t *st
         )
 {
     /* timings */
@@ -1755,7 +1822,7 @@ static void exact_sparse_linear_algebra_ff_8(
      * coefficients of new pivot rows */
     mat->cf_8  = realloc(mat->cf_8,
             (unsigned long)mat->nrl * sizeof(cf8_t *));
-    exact_sparse_reduced_echelon_form_ff_8(mat, bs, st);
+    exact_sparse_reduced_echelon_form_ff_8(mat, tbr, bs, st);
 
     /* timings */
     ct1 = cputime();
@@ -1765,15 +1832,16 @@ static void exact_sparse_linear_algebra_ff_8(
 
     st->num_zerored += (mat->nrl - mat->np);
     if (st->info_level > 1) {
-        printf("%7d new %7d zero", mat->np, mat->nrl - mat->np);
+        printf("%9d new %7d zero", mat->np, mat->nrl - mat->np);
         fflush(stdout);
     }
 }
 
 static void exact_sparse_dense_linear_algebra_ff_8(
         mat_t *mat,
+        const bs_t * const tbr,
         const bs_t * const bs,
-        stat_t *st
+        md_t *st
         )
 {
     len_t i;
@@ -1814,15 +1882,16 @@ static void exact_sparse_dense_linear_algebra_ff_8(
 
     st->num_zerored += (mat->nrl - mat->np);
     if (st->info_level > 1) {
-        printf("%7d new %7d zero", mat->np, mat->nrl - mat->np);
+        printf("%9d new %7d zero", mat->np, mat->nrl - mat->np);
         fflush(stdout);
     }
 }
 
 static void probabilistic_sparse_dense_linear_algebra_ff_8_2(
         mat_t *mat,
+        const bs_t * const tbr,
         const bs_t * const bs,
-        stat_t *st
+        md_t *st
         )
 {
     len_t i;
@@ -1863,15 +1932,16 @@ static void probabilistic_sparse_dense_linear_algebra_ff_8_2(
 
     st->num_zerored += (mat->nrl - mat->np);
     if (st->info_level > 1) {
-        printf("%7d new %7d zero", mat->np, mat->nrl - mat->np);
+        printf("%9d new %7d zero", mat->np, mat->nrl - mat->np);
         fflush(stdout);
     }
 }
 
 static void probabilistic_sparse_dense_linear_algebra_ff_8(
         mat_t *mat,
+        const bs_t * const tbr,
         const bs_t * const bs,
-        stat_t *st
+        md_t *st
         )
 {
     len_t i;
@@ -1910,7 +1980,7 @@ static void probabilistic_sparse_dense_linear_algebra_ff_8(
 
     st->num_zerored += (mat->nrl - mat->np);
     if (st->info_level > 1) {
-        printf("%7d new %7d zero", mat->np, mat->nrl - mat->np);
+        printf("%9d new %7d zero", mat->np, mat->nrl - mat->np);
         fflush(stdout);
     }
 }
@@ -1918,7 +1988,7 @@ static void probabilistic_sparse_dense_linear_algebra_ff_8(
 static void interreduce_matrix_rows_ff_8(
         mat_t *mat,
         bs_t *bs,
-        stat_t *st,
+        md_t *st,
         int free_basis
         )
 {
@@ -1929,8 +1999,18 @@ static void interreduce_matrix_rows_ff_8(
 
     /* adjust displaying timings for statistic printout */
     if (st->info_level > 1) {
-        printf("                        ");
+        printf("                          ");
     }
+
+    /* for interreduction steps like the final basis reduction we
+    need to allocate memory for rba here, even so we do not use
+    it at all */
+    mat->rba  = (rba_t **)malloc((unsigned long)ncols * sizeof(rba_t *));
+    const unsigned long len = ncols / 32 + ((ncols % 32) != 0);
+    for (i = 0; i < ncols; ++i) {
+        mat->rba[i] = (rba_t *)calloc(len, sizeof(rba_t));
+    }
+
     mat->tr = realloc(mat->tr, (unsigned long)ncols * sizeof(hm_t *));
 
     mat->cf_8  = realloc(mat->cf_8,
@@ -1955,6 +2035,8 @@ static void interreduce_matrix_rows_ff_8(
         if (pivs[l] != NULL) {
             memset(dr, 0, (unsigned long)ncols * sizeof(int64_t));
             cfs = bs->cf_8[pivs[l][COEFFS]];
+            const len_t bi  = pivs[l][BINDEX];
+            const len_t mh  = pivs[l][MULT];
             const len_t os  = pivs[l][PRELOOP];
             const len_t len = pivs[l][LENGTH];
             const hm_t * const ds = pivs[l] + OFFSET;
@@ -1972,16 +2054,20 @@ static void interreduce_matrix_rows_ff_8(
             pivs[l] = NULL;
             pivs[l] = mat->tr[k--] =
                 reduce_dense_row_by_known_pivots_sparse_ff_8(
-                        dr, mat, bs, pivs, sc, l, st->fc);
+                        dr, mat, bs, pivs, sc, l, mh, bi, 0, st->fc);
         }
     }
+    for (i = 0; i < ncols; ++i) {
+        free(mat->rba[i]);
+        mat->rba[i] = NULL;
+    }
     if (free_basis != 0) {
         /* free now all polynomials in the basis and reset bs->ld to 0. */
         free_basis_elements(bs);
     }    
     free(mat->rr);
     mat->rr = NULL;
-    mat->np = nrows;
+    st->np = mat->np = nrows;
     free(pivs);
     free(dr);
 }
